use std::collections::HashMap;

use crate::pos::Pos;

#[derive(Clone, PartialEq, Debug)]
pub struct Pack {
    stmts: Vec<Stmt>,
    exprs: Vec<Expr>,
    typs: Vec<Typ>,
    params: Vec<Param>,
    owners: HashMap<OwnerId, StmtId>,
}

impl<'pack> Pack {
    pub fn new(
        stmts: Vec<Stmt>,
        exprs: Vec<Expr>,
        typs: Vec<Typ>,
        params: Vec<Param>,
        owners: HashMap<OwnerId, StmtId>,
    ) -> Self {
        Self {
            stmts,
            exprs,
            typs,
            params,
            owners,
        }
    }

    pub fn stmt(&'pack self, id: StmtId) -> &'pack Stmt {
        &self.stmts[id.0]
    }

    pub fn expr(&'pack self, id: ExprId) -> &'pack Expr {
        &self.exprs[id.0]
    }

    pub fn typ(&'pack self, id: TypId) -> &'pack Typ {
        &self.typs[id.0]
    }

    pub fn param(&'pack self, id: ParamId) -> &'pack Param {
        &self.params[id.0]
    }
}

#[derive(Clone, Copy, PartialEq, Debug)]
pub struct StmtId(pub usize);

#[derive(Clone, Copy, PartialEq, Debug)]
pub struct ExprId(pub usize);

#[derive(Clone, Copy, PartialEq, Debug)]
pub struct TypId(pub usize);

#[derive(Clone, Copy, PartialEq, Debug)]
pub struct ParamId(pub usize);

#[derive(Clone, Copy, PartialEq, Eq, Hash, Debug)]
pub struct OwnerId(pub u64);

#[derive(Clone, PartialEq, Debug)]
pub struct Stmt {
    pub kind: StmtKind,
    pub pos: Pos,
}

impl Stmt {
    pub fn new(kind: StmtKind, pos: Pos) -> Self {
        Self { kind, pos }
    }
}

#[derive(Clone, PartialEq, Debug)]
pub enum StmtKind {
    Error,
    Let {
        subject: ParamId,
        value: ExprId,
    },
    Fn {
        subject_ident_id: u64,
        params: Vec<ParamId>,
        return_typ: Option<TypId>,
        body: ExprId,
    },
    Return {
        expr: Option<ExprId>,
    },
    Break {
        expr: Option<ExprId>,
    },
    Assign {
        subject: ExprId,
        value: ExprId,
    },
    Expr(ExprId),
}

#[derive(Clone, PartialEq, Debug)]
pub struct Expr {
    pub kind: ExprKind,
    pub pos: Pos,
}

impl Expr {
    pub fn new(kind: ExprKind, pos: Pos) -> Self {
        Self { kind, pos }
    }
}

#[derive(Clone, PartialEq, Debug)]
pub enum ExprKind {
    Error,
    Ident {
        ident_id: u64,
        owner: OwnerId,
    },
    Int(i64),
    Str(String),
    Group(ExprId),
    Block {
        stmts: Vec<StmtId>,
        expr: Option<ExprId>,
    },
    Call {
        subject: ExprId,
        args: Vec<ExprId>,
    },
    If {
        cond: ExprId,
        truthy: ExprId,
        falsy: Option<ExprId>,
    },
    Loop {
        body: ExprId,
    },
}

#[derive(Clone, PartialEq, Debug)]
pub struct Typ {
    pub kind: TypKind,
    pub pos: Pos,
}

impl Typ {
    pub fn new(kind: TypKind, pos: Pos) -> Self {
        Self { kind, pos }
    }
}

#[derive(Clone, PartialEq, Debug)]
pub enum TypKind {
    Error,
    Ident { ident_id: u64, owner: OwnerId },
}

#[derive(Clone, PartialEq, Debug)]
pub struct Param {
    pub subject_ident_id: u64,
    pub typ: Option<TypId>,
    pub pos: Pos,
}
