use std::{collections::HashMap, rc::Rc, sync::Mutex};

use crate::{
    ast::{Expr, ExprKind as EK, Param, Stmt, StmtKind as SK, Typ, TypKind},
    lexer::Lexer,
    pos::{Error, ErrorAcc, Pos},
    token::{Token, TokenKind as TK, TokenValue},
};

pub struct Parser<'a> {
    lexer: Lexer<'a>,
    current: Option<Token>,
    error_acc: Rc<Mutex<ErrorAcc>>,
}

impl<'a> Parser<'a> {
    pub fn new(text: &'a str, error_acc: Rc<Mutex<ErrorAcc>>) -> Self {
        let mut lexer = Lexer::new(text, error_acc.clone());
        let current = lexer.next();
        Self {
            lexer,
            current,
            error_acc,
        }
    }

    pub fn parse(&mut self) -> Vec<Stmt> {
        self.parse_file()
    }

    pub fn symbols(self) -> HashMap<u64, String> {
        self.lexer.symbols()
    }

    fn parse_file(&mut self) -> Vec<Stmt> {
        let mut stmts = Vec::new();
        loop {
            match self.current {
                Some(_) => stmts.push(self.parse_stmt()),
                None => break stmts,
            }
        }
    }

    fn parse_stmt(&mut self) -> Stmt {
        use TK::*;
        let pos = self.pos();
        match self.curr_kind() {
            Some(Fn) => self.parse_fn(),
            Some(kind @ (LBrace | If | Loop)) => Stmt::new(
                SK::Expr(Box::new(match kind {
                    LBrace => self.parse_block(),
                    If => self.parse_if(),
                    Loop => self.parse_loop(),
                    _ => unreachable!(),
                })),
                pos,
            ),
            Some(kind) => {
                let stmt = match kind {
                    Let => self.parse_let(),
                    Return => self.parse_return(),
                    Break => self.parse_break(),
                    _ => match self.parse_assign() {
                        ParsedAssign::Assign(stmt) => stmt,
                        ParsedAssign::Expr(expr, pos) => Stmt::new(SK::Expr(Box::new(expr)), pos),
                    },
                };
                self.eat_semicolon();
                stmt
            }
            _ => unreachable!(),
        }
    }

    fn parse_fn(&mut self) -> Stmt {
        let pos = self.pos();
        self.step();
        if !self.curr_is(TK::Ident) {
            self.error("expected id", pos.clone());
            return self.error_stmt(pos);
        }
        let subject = Box::new(self.parse_ident_expr());
        if !self.curr_is(TK::LParen) {
            self.error("expected '('", pos.clone());
            return self.error_stmt(pos);
        }
        let params = match self.parse_fn_params() {
            Ok(params) => params,
            Err(pos) => return self.error_stmt(pos),
        };
        let return_typ = if self.curr_is(TK::MinusLt) {
            self.step();
            Some(Box::new(self.parse_typ()))
        } else {
            None
        };
        if !self.curr_is(TK::LBrace) {
            self.error("expected '{'", pos.clone());
            return self.error_stmt(pos);
        }
        let body = Box::new(self.parse_block());
        Stmt::new(
            SK::Fn {
                subject,
                params,
                return_typ,
                body,
            },
            pos,
        )
    }

    fn parse_fn_params(&mut self) -> Result<Vec<Param>, Pos> {
        let pos = self.pos();
        self.step();
        let mut params = Vec::new();
        if !self.curr_is(TK::RParen) {
            if !self.curr_is(TK::Ident) {
                self.error("expected id", pos.clone());
                return Err(pos);
            }
            params.push(self.parse_param());
            while let Some(TK::Comma) = self.curr_kind() {
                self.step();
                if self.curr_is(TK::RParen) {
                    self.error("expected ')'", pos.clone());
                    break;
                }
                params.push(self.parse_param());
            }
        }
        if !self.curr_is(TK::RParen) {
            self.error("expected ')'", pos.clone());
            return Err(pos);
        }
        self.step();
        Ok(params)
    }

    fn parse_let(&mut self) -> Stmt {
        let pos = self.pos();
        self.step();
        if !self.curr_is(TK::Ident) {
            self.error("expected id", pos.clone());
            return self.error_stmt(pos);
        }
        let subject = self.parse_param();
        if !self.curr_is(TK::Equal) {
            self.error("expected '='", pos.clone());
            return self.error_stmt(pos);
        }
        self.step();
        let value = self.parse_expr();
        Stmt::new(
            SK::Let {
                subject: Box::new(subject),
                value: Box::new(value),
            },
            pos,
        )
    }

    fn parse_return(&mut self) -> Stmt {
        let pos = self.pos();
        self.step();
        let expr = match self.curr_kind() {
            Some(TK::Semicolon) => None,
            _ => Some(Box::new(self.parse_expr())),
        };
        Stmt::new(SK::Return { expr }, pos.clone())
    }

    fn parse_break(&mut self) -> Stmt {
        let pos = self.pos();
        self.step();
        let expr = match self.curr_kind() {
            Some(TK::Semicolon) => None,
            _ => Some(Box::new(self.parse_expr())),
        };
        Stmt::new(SK::Break { expr }, pos.clone())
    }

    fn parse_assign(&mut self) -> ParsedAssign {
        let pos = self.pos();
        let subject = self.parse_expr();
        match self.curr_kind() {
            Some(TK::Equal) => {
                self.step();
                let value = self.parse_expr();
                ParsedAssign::Assign(Stmt::new(
                    SK::Assign {
                        subject: Box::new(subject),
                        value: Box::new(value),
                    },
                    pos,
                ))
            }
            _ => ParsedAssign::Expr(subject, pos),
        }
    }

    fn parse_expr(&mut self) -> Expr {
        self.parse_call()
    }

    fn parse_call(&mut self) -> Expr {
        let pos = self.pos();
        let mut subject = self.parse_value();
        loop {
            match self.curr_kind() {
                Some(TK::LParen) => {
                    self.step();
                    let mut args = Vec::new();
                    match self.curr_kind() {
                        None | Some(TK::RParen) => {}
                        Some(_) => {
                            args.push(self.parse_expr());
                            while let Some(TK::Comma) = self.curr_kind() {
                                self.step();
                                if let Some(TK::RParen) = self.curr_kind() {
                                    break;
                                }
                                args.push(self.parse_expr());
                            }
                        }
                    }
                    match self.curr_kind() {
                        Some(TK::RParen) => {}
                        _ => {
                            self.error("expected ')'", pos.clone());
                            return self.error_expr(pos);
                        }
                    }
                    self.step();
                    subject = Expr::new(
                        EK::Call {
                            subject: Box::new(subject),
                            args,
                        },
                        pos.clone(),
                    );
                }
                _ => break subject,
            }
        }
    }

    fn parse_value(&mut self) -> Expr {
        let pos = self.pos();
        match self.curr_kind() {
            Some(TK::Ident) => self.parse_ident_expr(),
            Some(TK::Int) => self.parse_int(),
            Some(TK::Str) => self.parse_string(),
            Some(TK::LParen) => self.parse_group(),
            Some(TK::LBrace) => self.parse_block(),
            Some(TK::If) => self.parse_if(),
            Some(TK::Loop) => self.parse_loop(),
            _ => {
                self.error("expected value", pos.clone());
                self.step();
                self.error_expr(pos)
            }
        }
    }

    fn parse_ident_expr(&mut self) -> Expr {
        let pos = self.pos();
        let Some(Token {
            kind: TK::Ident,
            value: TokenValue::Ident(value),
            ..
        }) = self.current
        else {
            unreachable!()
        };
        self.step();
        Expr::new(EK::Ident(value), pos)
    }

    fn parse_int(&mut self) -> Expr {
        let pos = self.pos();
        let Some(Token {
            kind: TK::Int,
            value: TokenValue::Int(value),
            ..
        }) = self.current
        else {
            unreachable!()
        };
        self.step();
        Expr::new(EK::Int(value), pos)
    }

    fn parse_string(&mut self) -> Expr {
        let pos = self.pos();
        let Some(Token {
            kind: TK::Str,
            value: TokenValue::Str(value),
            ..
        }) = self.current.clone()
        else {
            unreachable!()
        };
        self.step();
        Expr::new(EK::Str(value.clone()), pos)
    }

    fn parse_group(&mut self) -> Expr {
        let pos = self.pos();
        self.step();
        let expr = Box::new(self.parse_expr());
        if !self.curr_is(TK::RParen) {
            self.error("expected ')'", pos.clone());
            return self.error_expr(pos);
        }
        self.step();
        Expr::new(EK::Group(expr), pos)
    }

    fn parse_block(&mut self) -> Expr {
        use TK::*;
        let pos = self.pos();
        self.step();
        let mut stmts = Vec::new();
        loop {
            match self.curr_kind() {
                None => {
                    self.error("expected '}'", pos.clone());
                    break self.error_expr(pos);
                }
                Some(RBrace) => {
                    self.step();
                    break Expr::new(EK::Block { stmts, expr: None }, pos);
                }
                Some(Fn) => stmts.push(self.parse_fn()),
                Some(kind @ (Let | Return | Break)) => {
                    let stmt = match kind {
                        Let => self.parse_let(),
                        Return => self.parse_return(),
                        Break => self.parse_break(),
                        _ => unreachable!(),
                    };
                    self.eat_semicolon();
                    stmts.push(stmt);
                }
                Some(kind @ (LBrace | If | Loop)) => {
                    let expr = match kind {
                        LBrace => self.parse_block(),
                        If => self.parse_if(),
                        Loop => self.parse_loop(),
                        _ => unreachable!(),
                    };
                    match self.curr_kind() {
                        Some(TK::Semicolon) => {
                            self.step();
                            stmts.push(Stmt::new(SK::Expr(Box::new(expr)), pos.clone()))
                        }
                        Some(TK::RBrace) => {
                            self.step();
                            break Expr::new(
                                EK::Block {
                                    stmts,
                                    expr: Some(Box::new(expr)),
                                },
                                pos,
                            );
                        }
                        _ => {}
                    }
                }
                Some(_) => match self.parse_assign() {
                    ParsedAssign::Assign(stmt) => {
                        self.eat_semicolon();
                        stmts.push(stmt);
                    }
                    ParsedAssign::Expr(expr, pos) => match self.curr_kind() {
                        Some(TK::Semicolon) => {
                            self.step();
                            stmts.push(Stmt::new(SK::Expr(Box::new(expr)), pos))
                        }
                        Some(TK::RBrace) => {
                            self.step();
                            break Expr::new(
                                EK::Block {
                                    stmts,
                                    expr: Some(Box::new(expr)),
                                },
                                pos,
                            );
                        }
                        _ => {}
                    },
                },
            }
        }
    }

    fn parse_if(&mut self) -> Expr {
        let pos = self.pos();
        self.step();
        let cond = Box::new(self.parse_expr());
        if !self.curr_is(TK::LBrace) {
            self.error("expected '}'", pos.clone());
            return self.error_expr(pos);
        }
        let truthy = Box::new(self.parse_block());
        let falsy = match self.curr_kind() {
            Some(TK::Else) => {
                self.step();
                if !self.curr_is(TK::LBrace) {
                    self.error("expected '}'", pos.clone());
                    return self.error_expr(pos);
                }
                Some(Box::new(self.parse_block()))
            }
            _ => None,
        };
        Expr::new(
            EK::If {
                cond,
                truthy,
                falsy,
            },
            pos,
        )
    }

    fn parse_loop(&mut self) -> Expr {
        let pos = self.pos();
        self.step();
        if !self.curr_is(TK::LBrace) {
            self.error("expected '}'", pos.clone());
            return self.error_expr(pos);
        }
        let body = Box::new(self.parse_block());
        Expr::new(EK::Loop { body }, pos)
    }

    fn parse_typ(&mut self) -> Typ {
        let pos = self.pos();
        match self.curr_kind() {
            Some(TK::Ident) => self.parse_ident_typ(),
            _ => {
                self.error("expected type", pos.clone());
                self.step();
                self.error_typ(pos)
            }
        }
    }

    fn parse_ident_typ(&mut self) -> Typ {
        let pos = self.pos();
        let Some(Token {
            kind: TK::Ident,
            value: TokenValue::Ident(value),
            ..
        }) = self.current
        else {
            unreachable!()
        };
        self.step();
        Typ::new(TypKind::Ident(value), pos)
    }

    fn parse_param(&mut self) -> Param {
        let pos = self.pos();
        let subject = Box::new(self.parse_ident_expr());
        let typ = if let Some(TK::Colon) = self.curr_kind() {
            self.step();
            Some(Box::new(self.parse_typ()))
        } else {
            None
        };
        Param { subject, typ, pos }
    }

    fn eat_semicolon(&mut self) {
        let pos = self.pos();
        match self.curr_kind() {
            Some(TK::Semicolon) => {
                self.step();
            }
            _ => {
                self.error("expected ';'", pos.clone());
            }
        }
    }

    fn error<S: Into<String>>(&mut self, msg: S, pos: Pos) {
        let msg = msg.into();
        self.error_acc.lock().unwrap().add(Error {
            kind: crate::pos::ErrorKind::Parser,
            pos: Some(pos),
            msg,
        });
    }

    fn step(&mut self) {
        self.current = self.lexer.next();
    }

    fn error_stmt(&self, pos: Pos) -> Stmt {
        Stmt::new(SK::Error, pos)
    }

    fn error_expr(&self, pos: Pos) -> Expr {
        Expr::new(EK::Error, pos)
    }

    fn error_typ(&self, pos: Pos) -> Typ {
        Typ::new(TypKind::Error, pos)
    }

    fn pos(&self) -> Pos {
        self.current
            .as_ref()
            .map(|token| token.pos.clone())
            .unwrap_or_else(|| self.lexer.pos())
    }

    fn curr_is(&self, kind: TK) -> bool {
        self.curr_kind() == Some(kind)
    }

    fn curr_kind(&self) -> Option<TK> {
        self.current.as_ref().map(|t| t.kind.clone())
    }
}

enum ParsedAssign {
    Assign(Stmt),
    Expr(Expr, Pos),
}
